'' System Version 2.001
'' OS_dspText Changes:
''
''   2006/09/24 - Initial version with change dates
''   2006/10/06 - Add TV & lores VGA compatibility routines
''   2006/10/26 - Added constants for display type and screen size

'' This is the VGA high resolution text driver for the Propeller OS.
'' It is based on the VGA hires text driver version 1.1  2006 by Parallax,
'' Inc.  Modifications  2006 by Michael Green.  Permission is given to
'' include this or portions of this program in any program for the
'' Parallax Propeller processor as long as proper credit is given.
'' This object generates a 1024x768 VGA signal which contains 128 columns x 64
'' rows of 8x12 characters. Each row can have a unique forground/background
'' color combination and each character can be inversed. There are also two
'' cursors which can be independently controlled (ie. mouse and keyboard). A
'' sync indicator signals each time the screen is refreshed (you may ignore).
''
'' You must provide buffers for the screen, colors, cursors, and sync. Once
'' started, all interfacing is done via memory. To this object, all buffers are
'' read-only, with the exception of the sync indicator which gets written with
'' -1. You may freely write all buffers to affect screen appearance. Have fun!

CON
{
' 1024 x 768 @ 57Hz settings: 128 x 64 characters

  hp = 1024     'horizontal pixels
  vp = 768      'vertical pixels
  hf = 16       'horizontal front porch pixels
  hs = 96       'horizontal sync pixels
  hb = 176      'horizontal back porch pixels
  vf = 1        'vertical front porch lines
  vs = 3        'vertical sync lines
  vb = 28       'vertical back porch lines
  hn = 1        'horizontal normal sync state (0|1)
  vn = 1        'vertical normal sync state (0|1)
  pr = 60       'pixel rate in MHz at 80MHz system clock (5MHz granularity)
}
{
' 800 x 600 @ 75Hz settings: 100 x 50 characters

  hp = 800      'horizontal pixels
  vp = 600      'vertical pixels
  hf = 40       'horizontal front porch pixels
  hs = 128      'horizontal sync pixels
  hb = 88       'horizontal back porch pixels
  vf = 1        'vertical front porch lines
  vs = 4        'vertical sync lines
  vb = 23       'vertical back porch lines
  hn = 0        'horizontal normal sync state (0|1)
  vn = 0        'vertical normal sync state (0|1)
  pr = 50       'pixel rate in MHz at 80MHz system clock (5MHz granularity)
}
'{
' 640 x 480 @ 69Hz settings: 80 x 40 characters

  hp = 640      'horizontal pixels
  vp = 480      'vertical pixels
  hf = 24       'horizontal front porch pixels
  hs = 40       'horizontal sync pixels
  hb = 128      'horizontal back porch pixels
  vf = 9        'vertical front porch lines
  vs = 3        'vertical sync lines
  vb = 28       'vertical back porch lines
  hn = 1        'horizontal normal sync state (0|1)
  vn = 1        'vertical normal sync state (0|1)
  pr = 30       'pixel rate in MHz at 80MHz system clock (5MHz granularity)
'}

' columns and rows

  cols         = hp / 8
  rows         = vp / 12
  screensize   = cols * rows
  TV           = false
  loresVGA     = false
  hiresVGA     = true

OBJ
  ldr : "OS_loader"
  
''   screenPtr = Pointer to 8,192 bytes containing ASCII codes for each of the
''               128x64 screen characters. Each byte's top bit controls color
''               inversion while the lower seven bits provide the ASCII code.
''               Screen memory is arranged left-to-right, top-to-bottom.
''
''               screen byte example: %1_1000001 = inverse "A"
''
''    colorPtr = Pointer to 64 words which define the foreground and background
''               colors for each row. The lower byte of each word contains the
''               foreground RGB data for that row, while the upper byte
''               contains the background RGB data. The RGB data in each byte is
''               arranged as %RRGGBB00 (4 levels each).
''
''               color word example: %%0020_3300 = gold on blue
''
''   cursorPtr = Pointer to 6 bytes which control the cursors:
''
''               bytes 0,1,2: X, Y, and MODE of cursor 0
''               bytes 3,4,5: X, Y, and MODE of cursor 1
''
''               X and Y are in terms of screen characters
''               (left-to-right, top-to-bottom)
''
''               MODE uses three bottom bits:
''
''                      %x00 = cursor off
''                      %x01 = cursor on
''                      %x10 = cursor on, blink slow
''                      %x11 = cursor on, blink fast
''                      %0xx = cursor is solid block
''                      %1xx = cursor is underscore
''
''               cursor example: 127, 63, %010 = blinking block in lower-right
''
''    videoPtr = Pointer to work area for Hires VGA driver.  See below for specifics
CON
  dsp_sync    =  0                             ' Sync.  Set to -1 at start of refresh
  dsp_cog     =  4                             ' Video cog bitmask      
  dsp_col     =  8                             ' Current column (screen relative)
  dsp_row     =  9                             ' Current row (screen relative)
  dsp_states  = 10                             ' Current state flags ("lazy" eol, insert)
  dsp_wleft   = 11                             ' Current window column offset
  dsp_wtop    = 12                             ' Current window top offset
  dsp_wcols   = 13                             ' Current window # cols
  dsp_wrows   = 14                             ' Current window # rows
  dsp_mode    = 15                             ' %ab_nnnn  a = update cursor 1 X,Y
                                               '           b = update cursor 0 X,Y
                                               '           n = current window number
  dsp_pcnt    = 16                             ' # of parameters to collect
  dsp_pptr    = 17                             ' Index of current parameter byte
  dsp_flag    = 18                             ' Saved "command" character
  dsp_params  = 19                             ' Buffer for parameters (5)
  dsp_wsave   = 24                             ' Save area for window state (128)
  
  dsp_LazyEol = %00000001                      ' dsp_status: at right end of window
  dsp_Insert  = %00000010                      ' dsp_status: insert mode (for editor)

  updateMask  = %001_0000                      ' Used to isolate cursor update bits

  winInfoSize = dsp_mode - dsp_col + 1         ' # bytes used for window state

'' Control codes.  For convenience, supplying a # cols and # rows of zero will use the maximum
'' allowable values (screen width - x, screen height - y).  Because of the limitations on colors
'' (foreground and background for each row only), the Color command will set the colors for a
'' specific row.  The colors do not scroll with the text.

  ClrWin      = $00                            ' clear window  
  Home        = $01                            ' home in window
  MoveXY      = $02                            ' move to X,Y in window (X and Y follow)
  CursLt      = $03                            ' cursor left
  CursRt      = $04                            ' cursor right
  CursUp      = $05                            ' cursor up
  CursDn      = $06                            ' cursor down
' Color       = $07                            '
  Bsp         = $08                            ' backspace
  Tab         = $09                            ' tab (8 spaces per)
  Lf          = $0A                            ' linefeed
  ClrEol      = $0B                            ' clear to end of line
' ClrEow      = $0C                            ' clear to end of active window
  Cr          = $0D                            ' carriage return
  MoveX       = $0E                            ' move to X in window (X follows)
  MoveY       = $0F                            ' move to Y in window (Y follows)
  SetWin      = $10                            ' set bounds of window (1-15,x,y,cols,rows)
  UseWin      = $11                            ' change to selected window (0-15)
' SetColor    = $12                            ' change color (0-15) to FG, BG (C, FG, BG follow)
  Color       = $13                            ' change the colors of a row (Y, FG, BG follow)
  Insert      = $14                            ' insert (move line to right one column)
  Cursor      = $15                            ' set cursor (%nuxyy follows)
                                               ' n = cursor #, u = underline(1)/solid(0)
                                               ' yy = off(00)/on(01)/slow(10)/fast(11)
  ScrollUp    = $16                            ' scroll up the current window
  ScrollDn    = $17                            ' scroll down the current window
  WinColor    = $18                            ' change window color to C (C follows)
  EscChr      = $1F                            ' (ESC) print next character C as-is (C follows)
  Delete      = $7F                            ' delete current character
  Alt00       = $FF                            ' changed to $00 by out (for use in string())
  ClrScr      = Alt00                          ' equivalent to ClrWin (for use in string())

  absX        = %0000                          ' X value is relative to screen coordinates             
  relX        = %0001                          ' X value is relative to current values
  winX        = %0010                          ' X value is relative to window coordinates
  absY        = %0000                          ' Y value is relative to screen coordinates
  relY        = %0100                          ' Y value is relative to current values
  winY        = %1000                          ' Y value is relative to window coordinates

  NoChgColor  = 16                             ' color # to indicate no change to current color
  UseWinColor = 17                             ' color # to cause use of current window color

PUB out(c) | i, j, k, p, f, x, y, t
'' Output a character.  This routine decodes the character stream, interpreting control codes.
'' These mostly follow the codes used with other text drivers with the exception of color since
'' hires vga color is very limited and associated with an entire screen row.

  p := long[ldr#videoPtr]
  if c == $FF
    c~
  if f := byte[p+dsp_flag]
    byte[p+dsp_params][byte[p+dsp_pptr]++] := c
    if byte[p+dsp_pptr] == byte[p+dsp_pcnt]
      case f
        MoveXY:
          updateCursor(byte[p+dsp_params][0],c,winX|winY,false)
          byte[p+dsp_states] &= !dsp_LazyEol
        SetWin:
          i := byte[p+dsp_params][0]
          if i > 0 and i < 16
            i := i * winInfoSize + p + dsp_wsave - dsp_col
            x := byte[p+dsp_params][1] <# cols - 1
            byte[i+dsp_col] := byte[i+dsp_wleft] := x
            y := byte[p+dsp_params][2] <# rows - 1
            byte[i+dsp_row] := byte[i+dsp_wtop] := y            
            if (t := byte[p+dsp_params][3] <# cols - x) == 0
              t := cols - x
            byte[i+dsp_wcols] := t
            if (t := c <# rows - y) == 0
              t := rows - y
            byte[i+dsp_wrows] := t
            byte[p+dsp_states] &= !dsp_LazyEol
            byte[i+dsp_mode]~
        UseWin:
          if c < 16
            i := (byte[p+dsp_mode] & %1111) * winInfoSize + p + dsp_wsave
            j := c * winInfoSize + p + dsp_wsave
            bytemove(i,p+dsp_col,winInfoSize)
            bytemove(p+dsp_col,j,winInfoSize)
            byte[p+dsp_mode] := (byte[p+dsp_mode] & !%1111) | c
            updateCursor(0,0,relX|relY,false)
        Color:
          y := byte[p+dsp_params][0] <# rows - 1
          word[long[ldr#colorPtr]][y] := (c << 8) | byte[p+dsp_params][1]
        MoveX:
          updateCursor(c,0,winX|relY,false)
          byte[p+dsp_states] &= !dsp_LazyEol
        MoveY:
          updateCursor(0,c,relX|winY,false)
        Cursor:
          i := c >> 4 & 1
          j := updateMask << i
          byte[long[ldr#cursorPtr+3*i]][2] := c & %111
          byte[p+dsp_mode] := (byte[p+dsp_mode] & !j) | ((c & %1000 > 0) & j)
          updateCursor(0,0,relX|relY,false)
        WinColor:
          ' For compatibility only.  Window color is ignored.
        EscChr:
          print(c)
      byte[p+dsp_flag]~
  else
    case c & $FF
      ClrWin:
        y := byte[p+dsp_wtop]
        t := byte[p+dsp_wrows]
        repeat i from y to y + t - 1
          fillwith(byte[p+dsp_wleft],i,byte[p+dsp_wcols]," ")
        updateCursor(0,0,winX|winY,false)
        byte[p+dsp_states] &= !dsp_LazyEol
      Home:     
        updateCursor(0,0,winX|winY,false)
        byte[p+dsp_states] &= !dsp_LazyEol
      MoveXY:
        byte[p+dsp_flag] := c
        byte[p+dsp_pptr]~
        byte[p+dsp_pcnt] := 2
      CursLt:
        updateCursor(-1,0,relX|relY,false)
        byte[p+dsp_states] &= !dsp_LazyEol
      CursRt:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          updateCursor(+1,0,relX|relY,true)
      CursUp:
        updateCursor(0,-1,relX|relY,false)
      CursDn:
        updateCursor(0,+1,relX|relY,false)
      SetWin:
        byte[p+dsp_flag] := c
        byte[p+dsp_pptr]~
        byte[p+dsp_pcnt] := 5
      Bsp:
        x := byte[p+dsp_col] - byte[p+dsp_wleft]
        y := byte[p+dsp_row] - byte[p+dsp_wtop]
        if x | y
          if x
            x--
          else
            x := byte[p+dsp_wcols] - 1
            y--
          updateCursor(x,y,winX|winY,false)
          t := byte[p+dsp_row] * cols + byte[p+dsp_col]
          byte[long[ldr#screenPtr] + t] := " "
      Tab:
        repeat
          print(" ")
        while (byte[p+dsp_col] - byte[p+dsp_wleft]) & 7
      Lf:
        if (byte[p+dsp_row] - byte[p+dsp_wtop]) == byte[p+dsp_wrows] - 1
          x := byte[p+dsp_col]
          y := byte[p+dsp_row]
          newline
          updateCursor(x,y,absX|absY,false)
        else
          updateCursor(0,+1,relX|relY,false)
      ClrEol:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          cleartoeol
      Color:
        byte[p+dsp_flag] := c
        byte[p+dsp_pptr]~
        byte[p+dsp_pcnt] := 3
      Cr:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          cleartoeol
        newline
      MoveX, MoveY, UseWin, Cursor, WinColor, EscChr:
        byte[p+dsp_flag] := c
        byte[p+dsp_pptr]~
        byte[p+dsp_pcnt] := 1
      Insert:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          i := long[ldr#screenPtr] + byte[p+dsp_row] * cols
          j := byte[p+dsp_col]
          k := byte[p+dsp_wleft] + byte[p+dsp_wcols] - 1
          repeat t from 0 to k - j
            byte[i + k - t] := byte[i + k - t - 1]
          byte[i + j] := " "
      Delete:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          i := long[ldr#screenPtr] + byte[p+dsp_row] * cols
          j := i + byte[p+dsp_col]
          k := byte[p+dsp_wleft] + byte[p+dsp_wcols] - 1
          bytemove(j,j+1,k - byte[p+dsp_col])
          byte[i+k] := " "
      ScrollUp:
        scrollUpWin
      ScrollDn:
        scrollDownWin
      other:
        print(c)

PUB str(stringptr)

'' Print a zero-terminated string

  repeat strsize(stringptr)
    out(byte[stringptr++])

PUB dec(value) | i

'' Print a decimal number

  if value < 0
    -value
    out("-")
  i := 1_000_000_000
  repeat 10
    if value => i
      out(value / i + "0")
      value //= i
      result~~
    elseif result or i == 1
      out("0")
    i /= 10

PUB decFixed(value,width) | i, p, f, a, b, c
'' Print a decimal number of fixed width
  p := @a
  f~
  if value < 0
    -value
    byte[p++] := "-"
  i := 1_000_000_000
  repeat 10
    if value => i
      byte[p++] := value / i + "0"
      value //= i
      f~~
    elseif f or i == 1
      byte[p++] := "0"
    i /= 10
  byte[p]~
  i := strsize(@a)
  if i < width
    repeat width - i
      out(" ")
  str(@a)
  
PUB hex(value, digits)

'' Print a hexadecimal number

  value <<= (8 - digits) << 2
  repeat digits
    out(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PUB bin(value, digits)

'' Print a binary number

  value <<= 32 - digits
  repeat digits
    out((value <-= 1) & 1 + "0")

PUB screenChar(x, y) | p
'' Return the character at the absolute screen coordinates x,y
  p := long[ldr#videoPtr]
  if x => cols or y => rows
    return 0
  return byte[y * cols + x + long[ldr#screenPtr]]

PUB screenColor(x, y) | p, s
'' Return the color of the character at absolute screen coordinates x,y
'' Because the hires VGA driver associates a color with a row, not with
'' individual characters, the color is always reported as UseWinColor.
  p := long[ldr#videoPtr]
  if x => cols or y => rows
    return NoChgColor
  return UseWinColor

PUB setScreenChar(x, y, c, z) | p
'' Set the character at the absolute screen coordinates x,y to c with color z
'' Because the hires VGA driver associates a color with a row, not with
'' individual characters, the color is ignored (but required for compatibility).
  p := long[ldr#videoPtr]
  if x => cols or y => rows
    return
  byte[y * cols + x + long[ldr#screenPtr]] := c

PUB windowChar(x, y) | p
'' Return the character at the window relative coordinates x,y
  p := long[ldr#videoPtr]
  if x => byte[p+dsp_wcols] or y => byte[p+dsp_wrows]
    return " "
  x += byte[p+dsp_wleft]
  y += byte[p+dsp_wtop]
  return byte[y * cols + x + long[ldr#screenPtr]]

PUB windowColor(x, y) | p
'' Return the color of the character at window relative coordinates x,y
'' Because the hires VGA driver associates a color with a row, not with
'' individual characters, the color is always reported as UseWinColor.
  p := long[ldr#videoPtr]
  if x => byte[p+dsp_wcols] or y => byte[p+dsp_wrows]
    return NoChgColor
  x += byte[p+dsp_wleft]
  y += byte[p+dsp_wtop]
  return screenColor(x,y)

PUB setWindowChar(x, y, c, z) | p
'' Set the character at window relative coordinates x,y to c with color z
'' Because the hires VGA driver associates a color with a row, not with
'' individual characters, the color is ignored (but required for compatibility).
  p := long[ldr#videoPtr]
  if x => byte[p+dsp_wcols] or y => byte[p+dsp_wrows]
    return
  x += byte[p+dsp_wleft]
  y += byte[p+dsp_wtop]
  setScreenChar(x,y,c,z)

PUB windowAtEnd(x, y) | p
'' Return true if window relative coordinates are past boundary
  p := long[ldr#videoPtr]
  return x => byte[p+dsp_wcols] or y => byte[p+dsp_wrows]

PUB windowCol | p
'' Return the current window relative x coordinate (column)
  p := long[ldr#videoPtr]
  return byte[p+dsp_col] - byte[p+dsp_wleft]

PUB windowRow | p
'' Return the current window relative y coordinate (row)
  p := long[ldr#videoPtr]
  return byte[p+dsp_row] - byte[p+dsp_wtop]

PUB out2(a, b)
'' Output a 2 character sequence
  out(a)
  out(b)

PUB out3(a, b, c)
'' Output a 3 character sequence
  out(a)
  out(b)
  out(c)

PUB out4(a, b, c, d)
'' Output a 4 character sequence
  out(a)
  out(b)
  out(c)
  out(d)

PUB out5(a, b, c, d, e)
'' Output a 5 character sequence
  out(a)
  out(b)
  out(c)
  out(d)
  out(e)

PUB out6(a, b, c, d, e, f)
'' Output a 6 character sequence
  out(a)
  out(b)
  out(c)
  out(d)
  out(e)
  out(f)

PRI print(c) | p
  p := long[ldr#videoPtr]
  if (byte[p+dsp_states] & dsp_LazyEol) <> 0
    newline
  byte[long[ldr#screenPtr] + byte[p+dsp_row] * cols + byte[p+dsp_col]] := c
  updateCursor(+1,0,relX|relY,true)

PRI updateCursor(x,y,mode,chkEol) | p, ax, ay, wx, wy, m, q, i
  p := long[ldr#videoPtr]
  ax := byte[p+dsp_wleft]
  ay := byte[p+dsp_wtop]
  wx := ax + byte[p+dsp_wcols] - 1
  wy := ay + byte[p+dsp_wrows] - 1
  case mode & %0011
    relX:
      x += byte[p+dsp_col]
      x &= x & $80000000 == 0
    winX:
      x += ax
  case (mode & %1100) >> 2
    relY >> 2:
      y += byte[p+dsp_row]
      y &= y & $80000000 == 0
    winY >> 2:
      y += ay
  if chkEol
    if x > wx or y > wy
      byte[p+dsp_states] |= dsp_LazyEol
    else
      byte[p+dsp_states] &= !dsp_LazyEol
  byte[p+dsp_col] := x := x #> ax <# wx
  byte[p+dsp_row] := y := y #> ay <# wy
  if m := byte[p+dsp_mode]
    q := long[ldr#cursorPtr]
    repeat i from 0 to 1
      if updateMask << i & m
        byte[3*i+q][0] := x
        byte[3*i+q][1] := y

PRI newline | p
  p := long[ldr#videoPtr]
  updateCursor(0,+1,winX|relY,true)
  if (byte[p+dsp_states] & dsp_LazyEol) <> 0
    scrollUpWin
  byte[p+dsp_states] &= !dsp_LazyEol

PRI scrollUpWin | i, k, p, s, e
  p := long[ldr#videoPtr]
  s := long[ldr#screenPtr]
  if byte[p+dsp_wrows] > 1
    e := byte[p+dsp_wtop] + byte[p+dsp_wrows] - 1
    repeat i from byte[p+dsp_wtop] to e - 1
      moveline(byte[p+dsp_wleft],i,byte[p+dsp_wcols],i+1)
    k := byte[p+dsp_col]
    updateCursor(0,e,winX|absY,false)
    cleartoeol
    updateCursor(k,0,absX|relY,false)

PRI scrollDownWin | i, k, p, s, e
  p := long[ldr#videoPtr]
  s := long[ldr#screenPtr]
  if byte[p+dsp_wrows] > 1
    e := byte[p+dsp_wtop] + byte[p+dsp_wrows] - 1
    repeat i from e to byte[p+dsp_wtop] + 1
      moveline(byte[p+dsp_wleft],i,byte[p+dsp_wcols],i-1)
    k := byte[p+dsp_col]
    updateCursor(0,0,winX|winY,false)
    cleartoeol
    updateCursor(k,0,absX|winY,false)

PRI cleartoeol | p, x, w
 p := long[ldr#videoPtr]
 x := byte[p+dsp_col]
 w := byte[p+dsp_wleft] + byte[p+dsp_wcols] - 1
 fillwith(x,byte[p+dsp_row],w-x," ")
       
PRI fillwith(x, y, w, c) | i, s
  s := long[ldr#screenPtr] + y * cols
  repeat i from x to x + w - 1
    byte[s + i] := c

PRI moveline(x, y, w, f) | i, s, t
  s := long[ldr#screenPtr] + y * cols
  t := long[ldr#screenPtr] + f * cols
  repeat i from x to x + w - 1
    byte[s + i] := byte[t + i]