'' System Version 2.001  Modified by Nico Hattink for use in Master-Slave debugging
'' (Deleted all experimental I/O routines, added some keyboard commands)
'' All routines (except drivers where noted) are (c) 2006 by Michael Green.
'' Some I/O drivers are (c) by Parallax, some with modifications by others.
'' Permission is given to include this or portions of this program in any program
'' for the Parallax Propeller processor as long as proper credit is given.
CON
  _clkmode    = xtal1 + pll16x
  _xinfreq    = 5_000_000
OBJ
  init    : "OS_initialization"                ' OS Initialization routines and definitions
  ldr     : "OS_loader"                        ' SPIN program loader and support routines
  dsp     : "OS_dspText"                       ' Display text driver modified for OS use
  key     : "OS_keyboard"                      ' PS/2 keyboard driver modified for OS use
  files   : "OS_fileSystem"                    ' File system support
  edit    : "OS_windowEditor"                  ' Screen and file editor
 {------------------------------------------------------------------------}
  FREE    : "COG_counter"                       'counts free COG's
  mast    : "Rem-Con-Mas.spin"                  'remote control master

CON
  _free = (ldr#endMemory - ldr#endFree + ldr#totalNeeded + 3) / 4  ' # longs needed
  #1, processProbe, processInfo,  processEdit  ' For command post-processing
      processHelp,  processRun,   processList
      processCopy,  processDump,  processDelete
      processLoad
 {-----------------------------------------------------------------------------}
      NAH1, NAH2, NAH3, NAH4, NAH5, NAH6, NAH7, NAH8 
      NAH9, NAH10, NAH11, NAH12, NAH13, NAH14, NAH15, NAH16 
VAR
  long dumpPin, dumpAddr, dumpSize
  long fileAddr, fileAddrTo, fileChange
  long fileName[3], fileNameTo[3]

  long par1,par2,par3,par4                           'NAH keyboard input parameters
  long flags[16],slid                                'flags are shared data with slaves  
  byte message[80]                                    'message from slave
  
PUB processCommands | t, y, i, j, k, z, b0, b1, b2, b3, b4, b5, b6, b7, tel ,cnt1,cnt2,cnt3 
'' Initialize the operating system, then repeatedly request a one line
'' command from the user (in a single line window), then execute it in a
'' full screen window.
  init.start(string("Propeller OS - Version 2.001.04 Modified NAH for MS debug",dsp#Cr))
 {-------------------------------------------------------------------------}
  slid:=1                                      'default slave-id on startup       
  repeat
    y := dsp.windowRow
    t := \enterDoCommand(y)
    if t < 0
      dsp.out2(dsp#UseWin,0)
      dsp.out4(dsp#MoveXY,dsp#cols,y,dsp#Cr)
      dsp.str(string("*** "))
      dsp.str(-t)
      dsp.out(dsp#Cr)
    else
      dsp.out2(dsp#UseWin,0)
      dsp.out4(dsp#MoveXY,dsp#cols,y,dsp#Cr)
      case t
        processProbe:
          if dumpPin < (ldr#maxPinPairs << 1)
            if (\files.updateBitmapEEPROMs(dumpPin) & $4000) == 0
              dsp.str(string("*** no boot eeprom?",dsp#Cr))
              return
          repeat i from 0 to (ldr#maxPinPairs<<1)-2 step 2
            z~
            y := word[long[ldr#systemPtr]+ldr#sys_eeproms+i]
            if y <> 0
              repeat j from 0 to 15
                if (|<j & y) <> 0
                  z += 32
              dsp.str(string("    Pin "))
              dsp.decFixed(i,2)
              dsp.str(string("  "))
              dsp.decFixed(z,3)
              dsp.str(string("K",dsp#Cr))
        processInfo:
          dsp.str(string("    pbase="))
          ldr.readEEPROM(fileAddr+ldr#spin_pbase,@t,4)
          dsp.hex(t,4)
          dsp.str(string(", vbase="))
          ldr.readEEPROM(fileAddr+ldr#spin_vbase,@t,4)
          dsp.hex(t,4)
          dsp.str(string(", dbase="))
          ldr.readEEPROM(fileAddr+ldr#spin_dbase,@t,4)
          dsp.hex(t,4)
          dsp.str(string(dsp#Cr,"    pcurr="))
          ldr.readEEPROM(fileAddr+ldr#spin_pcurr,@t,4)
          dsp.hex(t,4)
          dsp.str(string(", dcurr="))
          ldr.readEEPROM(fileAddr+ldr#spin_dcurr,@t,4)
          dsp.hex(t,4)
          dsp.out(dsp#Cr)
        processEdit:
          editWindow
        processHelp:
'' Copy 32K (0-$7FEF) from one file to another.  Create a file if needed.
          dsp.str(string(" copy <fileFrom> <fileTo>",dsp#Cr))
'' Delete the specified file
          dsp.str(string(" delete <file>",dsp#Cr))
'' Display the specified EEPROM or HUB RAM locations
          dsp.str(string(" dump {<pin>} <addr> <len>",dsp#Cr))
'' Start a simple screen editor.  Quit using the App Key (to the left of Ins)
          dsp.str(string(" edit",dsp#Cr))
'' Display a list of accepted commands
          dsp.str(string(" help",dsp#Cr))
'' Show SPIN related pointers for the program in the specified file
          dsp.str(string(" info <file>",dsp#Cr))
'' List files available including their location and option bits
          dsp.str(string(" list",dsp#Cr))

'' Optionally probe the possible I2C buss on the specified pin pair for EEPROMs
'' List all known I2C busses and the total amount of EEPROM space on each
          dsp.str(string(" probe {<pin>}",dsp#Cr))
'' Execute the SPIN program contained in the specified file
          dsp.str(string(" run <file>",dsp#Cr))
        processList:
          files.fileList
        processDump:
          files.dumpMemory(dumpPin,dumpAddr,dumpSize)
        processCopy:
          if fileChange
            dsp.str(string("    Overwrite? (Y/n) "))
            repeat
              t := key.getkey
              if t == dsp#Cr
                t := "y"
            until t == "Y" or t == "y" or t == "N" or t == "n"
            dsp.out(t)
          else
            t := "y"
            dsp.str(string("Created"))
          if t == "Y" or t == "y"
            dsp.out(" ")
            j := dsp.windowCol
            repeat i from 0 to $7FE0 step 32
              dsp.out2(dsp#MoveX,j)
              dsp.decFixed(i,5)
              if i == $7FE0
                k := 16
              else
                k := 32
              t := cnt
              repeat until ldr.readEEPROM(fileAddr+i,@b0,k)
                if (cnt - t) > (clkfreq / 200)
                  abort -string("EEPROM read timeout")
              t := cnt
              repeat until ldr.writeEEPROM(fileAddrTo+i,@b0,k)
                if (cnt - t) > (clkfreq / 200)
                  abort -string("EEPROM write timeout")
            dsp.out2(dsp#MoveX,j)
            dsp.str(string("Saved"))
          dsp.out(dsp#Cr)
        processRun:
          dsp.str(string("    Loading ... ",dsp#Cr))
          repeat i from 0 to 7  ' Stop any cogs not registered
            if (word[long[ldr#systemPtr]+ldr#sys_cog] & |<i) == 0
              cogstop(i)
          ldr.bootEEPROM(fileAddr)
          dsp.str(string("    Error during loading",dsp#Cr))
        processDelete:
          dsp.str(string("    Are you sure? (Y/n) "))
          repeat
            t := key.getkey
            if t == dsp#Cr
              t := "y"
          until t == "Y" or t == "y" or t == "N" or t == "n"
          dsp.out(t)
          if t == "Y" or t == "y"
            files.markDeleted(@fileName,fileAddr)
            dsp.str(string(" - Done"))
          dsp.out(dsp#Cr)
{------------------------------------------------------------------------------------------------}
{---------------------------------------------------NAH added keyboard commands------------------}
        NAH1:                                           '"start" debug master Cog
         tel:=mast.start
         dsp.str(string("Debug master runs in COG no"))
         dsp.decFixed(tel-1,2)
         dsp.out(dsp#CR)
         
        NAH2:                                           '"stop <Cog>"
         cogstop(par1)        

        NAH3:                                           '"free", shows number of free Cog's
         dsp.str(string("  The number of free COG's is "))
         dsp.decFixed(free.free_cogs,4)
         dsp.out(dsp#CR)       

        NAH4:                                          '"slave <ID>", set ID of slave to talk with
         slid:=par1

        NAH5:                                          '"pins", show slave pin states
         repeat while not key.gotkey
            dsp.str(string("slave "))                             
            dsp.hex(slid,2)
            par1:=mast.readpin(slid)                   'read all pins from slave 
            repeat i from 3 to 0                       'process 4 bytes, byte 3 is MSB
              dsp.str(string(" A"))
              dsp.dec((i<<3)+7)
              tel:=par1.byte[i]
              dsp.out($20)      
              dsp.bin(tel,8)
            dsp.out(dsp#cr)
           waitcnt(50000000+CnT)        

        NAH8:                                         '"meml <address>", show memory contents long
          repeat while not key.gotkey                 'best for display of variables etc
            dsp.str(string("slave "))                 'in big endian notation  
            dsp.hex(slid,2)
            dsp.str(string(" address "))           
            dsp.hex(par1,4)
            repeat i from 0 to 3                       'read 4 hub memory locations from slave
              flags[i]:=mast.readlong(slid,par1)        
              dsp.out($20) 
              par1:=par1+4
             dsp.hex(flags[i],8)
           dsp.out(dsp#cr)
           waitcnt(50000000+CnT)       

        NAH7:                                           '"trace", shows slave trace info                                                              
          repeat while not key.gotkey
            dsp.str(string("slv"))
            dsp.hex(slid,2)
            repeat i from 0 to 3  
             j:=i<<1
             dsp.str(string(" TP"))
             dsp.hex(i,1)
             dsp.out ($20)
             dsp.hex(mast.readflag(slid,j),3)   
             dsp.out($20)
             dsp.hex(mast.readflag(slid,j+1),8)            
           dsp.out(dsp#cr)
           waitcnt(50000000+CnT)        

        NAH6:                                            '"memb <address>", display hub memory bytes
          repeat while not key.gotkey                    'best for strings and "binary" interpretation
            dsp.str(string("slave "))                    'in little endian representation     
            dsp.hex(slid,2)
            dsp.str(string(" address "))           
            dsp.hex(par1,4)
            repeat i from 0 to 7                         
              dsp.out($20) 
              dsp.hex(mast.readbyte(slid,par1++),2)        
            dsp.out(dsp#cr)
            waitcnt(50000000+CnT)        

        NAH9:                                              '"break", show breakpoint status
          dsp.str(string("slave "))   
          dsp.hex(slid,2)
          dsp.str(string(" is waiting on ")) 
          dsp.hex(mast.readflag(slid,9),4)
          dsp.str(string(" value "))
          dsp.hex(mast.readflag(slid,10),8)
          dsp.out(dsp#cr) 
                                                                      
       NAH10:                                                '"cont", continue after break  
          mast.writeflag(slid,9,0)                            'reset breakpoint

       NAH11:                                                '"mess", display messages from slave   
         repeat while not key.gotkey
          tel:=mast.readflag(slid,7)
          if tel <>  0
             repeat i from 0 to 79
              message[i]:=mast.readbyte(slid,tel++)
             dsp.str(@message)
           dsp.out(dsp#cr)
           mast.writeflag(slid,7,0)                           'reset, message

       NAH12: dsp.str(string(" No test available"))           'test, place for test routines
          dsp.out(dsp#cr) 

        
PUB editWindow | t
'' Set up a simple screen editor with numbered blank lines.  Use the "App" key
'' to exit the editor (this is left of the "Ins" key on the Parallax keyboard).
  dsp.out6(dsp#SetWin,1,5,1,dsp#cols-5,dsp#rows-2)
  dsp.out6(dsp#SetWin,2,0,1,5,dsp#rows-2)
  dsp.out5(dsp#UseWin,2,dsp#WinColor,0,dsp#ClrScr)
  repeat t from 1 to dsp#rows - 2
    dsp.out3(dsp#MoveXY,0,t-1)
    dsp.decFixed(t,4)
  dsp.out3(dsp#UseWin,1,dsp#ClrScr)
  repeat
    t := edit.windowEditor
  until t == $CC
  dsp.out3(dsp#UseWin,0,dsp#ClrScr)

PUB enterDoCommand(y) | t, x, a, b, c
'' Set up a window at the line specified and accept a command
'' line terminated by a Cr.  Scan for a command word and
'' command specific parameters, either words or numbers.
'' If the command is known and the parameters are valid,
'' execute the command.  Abort with a negative error string
'' address or return a positive command specific value.
  dsp.out6(dsp#SetWin,1,1,y,dsp#cols-1,1)
  dsp.out6(dsp#UseWin,0,dsp#MoveXY,0,y,">")
  dsp.out5(dsp#UseWin,1,dsp#WinColor,0,dsp#ClrScr)
  repeat
    t := edit.windowEditor
  until t == dsp#Cr
  x := \getTextWord(0,@a,4,true," ")
  if x < 0
    abort -string("no command given")
  case a
    $59504F43:                                           ' copy
      x := \getTextWord(x,@fileName,12,true," ")
      if x < 0
        abort -string("missing <fileFrom>")
      x := \getTextWord(x,@fileNameTo,12,true," ")
      if x < 0
        abort -string("missing <fileTo>")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<fileFrom> not found")
      fileChange := true
      fileAddrTo := \files.searchFile(@fileNameTo)
      if fileAddrTo < 0
        fileAddrTo := \files.searchFile(0)
        if fileAddrTo < 0
          abort -string("can't create <fileTo>")
        files.markNewName(@fileNameTo,fileAddrTo,0)
        fileChange := false
      return processCopy
    $424F5250:                                          ' probe
      x := \getNumber(x,@dumpPin)
      if x < 0
        dumpPin := ldr#maxPinPairs << 1
        return processProbe
      if (dumpPin < (ldr#maxPinPairs << 1)) and (dumpPin & 1 == 0)
        return processProbe
      else
        abort -string("pin number odd or does not exist")
    $204E5552:                                           ' run
      if (x := \getTextWord(x,@fileName,12,true," ")) < 0
        abort -string("missing <file>")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<file> not found")
      return processRun
    $54494445:                                           ' edit
      return processEdit
    $4F464E49:                                           ' info
      x := \getTextWord(x,@fileName,12,true," ")
      if x < 0
        abort -string("missing <file>")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<file> not found")
      return processInfo
    $504C4548:                                           ' help
      return processHelp
    $5453494C:                                           ' list
      return processList
    $454C4544:                                           ' delete
      x := \getTextWord(x,@fileName,12,true," ")
      if x < 0
        abort -string("missing <file>")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<file> not found")
      if fileAddr == (ldr#i2cBootSCL << 18)
        abort -string("can't delete boot EEPROM")
      return processDelete
    $504D5544:                                           ' dump
      x := \getNumber(x,@a)
      if x < 0
        abort -string("missing <pin> or <addr>")
      x := \getNumber(x,@b)
      if x < 0
        abort -string("missing <addr> or <len>")
      x := \getNumber(x,@c)
      if x < 0
        if (a > $7FFFF) or (b => 1024)
          abort -string("invalid number")
        dumpPin := ldr#maxPinPairs << 1
        dumpAddr := a
        dumpSize := b
        return processDump
      else
        if (a => (ldr#maxPinPairs << 1)) or (a & 1) or (b > $7FFFF) or (c => 1024)
          abort -string("invalid number")
        dumpPin := a
        dumpAddr := b
        dumpSize := c
        return processDump
{--------------------------------------------------------------------------------------------}
{---------------------------------------NAH added keyboard functions-------------------------}
    $52415453:                                          'start              NAH1
      return NAH1

   $504f5453:                                           'stop <fie>         NAH2
      x := \getNumber(x,@par1)
      if x < 0
        abort -string("Cog number not found")
      return NAH2

   $45455246:                                           'free               NAH3
       return NAH3

   $56414c53:                                           'slave <number>     NAH4
      x := \getNumber(x,@par1)                          
      if x < 0
         abort -string("no slave-id given")
    return NAH4   

   $534e4950:                                           'pins               NAH5
        return NAH5  

   $424d454d:                                           'memb <address>      NAH6
       x := \getNumber(x,@par1)                         'byte ordered memory display
      if x < 0
         abort -string("no address given")
      return NAH6  

   $43415254:                                           'trace               NAH7
     return NAH7   

   $4c4d454d:                                           'meml <address>      NAH8
       x := \getNumber(x,@par1)                         '
      if x < 0
         abort -string("no address given")
      return NAH8  

   $41455242:                                           'break, show breakpoint status   NAH9
      return NAH9  

   $544e4f43:                                           'cont, continue after break      NAH10
      return NAH10   

   $5353454d:                                           'mess, show message from slave 
     return NAH11

   $54534554:                                           'test, for temporary test routines
     return NAH12

   other:
      abort -string("unknown command")

PRI skipBlanks(x)
'' Skip blanks until at right border of window and return x
'' of first non-blank or abort with a negative error string
'' address.
  repeat
    if dsp.windowAtEnd(x,0)
      abort -string("missing non-blank")
  until dsp.windowChar(x++,0) <> " "
  return x - 1

PRI getTextWord(x,textAddr,maxSize,uppercase,clearFirst) | c
'' Skip leading blanks, then copy up to maxSize non-blank
'' characters into the area at textAddr.  Return the
'' position of the next blank.  If uppercase is true,
'' store any lower case alphabetic characters as upper
'' case.  If clearFirst is not -1 (true), clear the
'' textAddr area to the value given.  If an error occurs,
'' abort with a negative error string address
  if clearFirst <> -1
    repeat c from 1 to maxSize
      byte[textAddr + c - 1] := clearFirst
  x := skipBlanks(x)
  repeat while (c := dsp.windowChar(x++,0)) <> " "
    if maxSize > 0
      if uppercase and c => "a" and c =< "z"
        c := c & !%0100000
      byte[textAddr++] := c
      maxSize--
  return x - 1

PRI getNumber(x,numberAddr) | negative, c, t
'' Skip leading blanks, then convert a decimal number
'' with an optional sign ("+"/"-") into binary to be
'' stored in the long at numberAddr.  Return the
'' position of the next blank.  If an error occurs,
'' abort with a negative error string address.
'' Currently, there's no checking for 32 bit overflow.
  x := skipBlanks(x)
  c := dsp.windowChar(x++,0)
  negative := c == "-"
  if negative or c == "+"
    c := dsp.windowChar(x++,0)
  if c == "$"
    c := dsp.windowChar(x++,0)
    t := hexDigit(c)
    if t < 0
      abort -string("invalid hex character")
    long[numberAddr] := t
    c := dsp.windowChar(x++,0)
    repeat until (t := hexDigit(c)) < 0
      long[numberAddr] := long[numberAddr] << 4 | t
      c := dsp.windowChar(x++,0)
    if negative
      long[numberAddr] := -long[numberAddr]
    return x - 1
  if c == "%"
    c := dsp.windowChar(x++,0)
    if not (c == "0" or c == "1")
      abort -string("invalid hex character")
    long[numberAddr] := c - "0"
    c := dsp.windowChar(x++,0)
    repeat while c == "0" or c == "1"
      long[numberAddr] := long[numberAddr] << 1 | (c - "0")
      c := dsp.windowChar(x++,0)
    if negative
      long[numberAddr] := -long[numberAddr]
    return x - 1
  if not (c => "0" and c=< "9")
    abort -string("invalid numeric character")
  long[numberAddr] := c - "0"
  c := dsp.windowChar(x++,0)
  repeat while c => "0" and c =< "9"
    long[numberAddr] := long[numberAddr] * 10 + c - "0"
    c := dsp.windowChar(x++,0)
  if negative
    long[numberAddr] := -long[numberAddr]
  return x - 1

PRI hexDigit(c)
'' Convert a hexadecimal character to the corresponding value or -1 if invalid.
  if c => "0" and c =< "9"
    return c - "0"
  if c => "A" and c =< "F"
    return c - "A" + 10
  if c => "a" and c =< "f"
    return c - "a" + 10
  return -1
  